import pytest
import tempfile
import os
import re
import sys
from logging import warning

from .utils.persistent_temporary_directory import PersistentTemporaryDirectory
if PersistentTemporaryDirectory: # Supress pyright unused symbol warnings
    pass

from .utils.misc import current_test_name, copy_dict_to_fs


path_to_source_modules=os.path.abspath(os.path.join(os.path.dirname(__file__), '../src'))
sys.path.insert(0, path_to_source_modules)

import continuing_education
from continuing_education import  RemoteCheckpointSynchronizer
from continuing_education.checkpoint_registry import S3RemoteCheckpointSynchronizer
continuing_education.checkpoint_registry.verbose = False # Make this True for debugging

bucket_name='test_bucket'
prefix='/checkpoints/test'
s3_uri = f's3://{bucket_name}/{prefix}'
local_output_dir = 'test_outputs'

unsupported_uri = f'mongo://whatever/monog/uris/have/in/the/path'

def test_handler_registry_populated() -> None:
    assert len(RemoteCheckpointSynchronizer._handler_registry) > 0
    assert S3RemoteCheckpointSynchronizer in RemoteCheckpointSynchronizer._handler_registry

def test_can_handle_uri() -> None:
    assert S3RemoteCheckpointSynchronizer.can_handle_uri(s3_uri)

def test_from_uri_with_s3_uri() -> None:
    synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir)
    assert synchronizer.__class__ == S3RemoteCheckpointSynchronizer


def test_from_uri_with_unsupported_uri() -> None:
    with pytest.raises(RuntimeError) as err_info:
        synchronizer = RemoteCheckpointSynchronizer.from_uri(unsupported_uri, local_output_dir)
        assert not synchronizer
    assert 'No RemoteCheckpointSynchronizer exists that can handle the uri' in str(err_info.value)


def test_construct_with_real_uri(populated_s3_checkpoints):
    s3_uri = populated_s3_checkpoints['uri']
    if s3_uri:
        contents = populated_s3_checkpoints['contents']
        assert contents
        with tempfile.TemporaryDirectory() as temp_dir:
            synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir = temp_dir)
            assert synchronizer
    else:
        reason = populated_s3_checkpoints['reason']
        warning(f'Skipping {current_test_name()}() because {reason}')
        assert True


def test_download_one_sync_with_file_that_exists(populated_s3_checkpoints):
    s3_uri = populated_s3_checkpoints['uri']
    if s3_uri:
        contents = populated_s3_checkpoints['contents']
        rel_path = 'checkpoint_registry.json' if 'checkpoint_registry.json' in contents else contents.keys()[0]
        remote_file_content = contents[rel_path]
        with tempfile.TemporaryDirectory() as temp_dir:
            synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir = temp_dir)
            assert synchronizer
            synchronizer.download_one_sync(rel_path)
            full_path = os.path.join(temp_dir,rel_path)
            assert os.path.exists(full_path), f'{full_path} does not exist'
            local_file_content = open(full_path,'r').read()
            assert local_file_content == remote_file_content
    else:
        reason = populated_s3_checkpoints['reason']
        warning(f'Skipping {current_test_name()}() because {reason}')
        assert True

def test_download_one_sync_with_file_nonexistent(populated_s3_checkpoints):
    s3_uri = populated_s3_checkpoints['uri']
    if s3_uri:
        contents = populated_s3_checkpoints['contents']
        rel_path = 'nonexistent_file.bogus'
        print(f'checking for nonexistent file "{rel_path}"')
        assert not rel_path in contents
        #with PersistentTemporaryDirectory() as temp_dir:
        with tempfile.TemporaryDirectory() as temp_dir:
            print(f'the tempdir is {temp_dir}/')
            synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir = temp_dir)
            assert synchronizer
            synchronizer.download_one_sync(rel_path)
            full_path = os.path.join(temp_dir,rel_path)
            assert not os.path.exists(full_path), f'{full_path} should not exist because it was no on the remote'
    else:
        reason = populated_s3_checkpoints['reason']
        warning(f'Skipping {current_test_name()}() because {reason}')
        assert True

def confirm_directory_download_correct(temp_dir: str, rel_path: str, contents: dict[str,str]) -> bool:
    if rel_path == '':
        keys_under_rel_path = list(contents.keys())
    else:
        keys_under_rel_path = list(filter(lambda x: x.startswith(os.path.join(rel_path,'')), contents.keys()))

    # print(f'Confirming content of objects under {rel_path}: {keys_under_rel_path}')
    for key in keys_under_rel_path:
        corresponding_local_path = os.path.join(temp_dir,key)
        # print(f'comparing {key} with {corresponding_local_path}')
        assert os.path.exists(corresponding_local_path), f'{corresponding_local_path} does not exist'
        local_file_content = open(corresponding_local_path,'r').read()
        assert local_file_content == contents[key]
    return True

def test_download_one_sync_with_directory(populated_s3_checkpoints):
    s3_uri = populated_s3_checkpoints['uri']
    if s3_uri:
        contents = populated_s3_checkpoints['contents']

        first_matching_checkpoint_dash_number = None
        pattern = r'^(checkpoint-\d+)/.+$'
        for k in contents.keys():
            match = re.match(pattern,k)
            if match:
                first_matching_checkpoint_dash_number = match.group(1)
                break
        assert first_matching_checkpoint_dash_number, f'no checkpoint directory found in {s3_uri}'
        rel_path = first_matching_checkpoint_dash_number
        # print(f'trying download_one_sync on rel_path "{rel_path}"')
        with tempfile.TemporaryDirectory() as temp_dir:
#        with PersistentTemporaryDirectory() as temp_dir:
            synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir = temp_dir)
            assert synchronizer
            synchronizer.download_one_sync(rel_path)
            full_path = os.path.join(temp_dir,rel_path)
            assert os.path.exists(full_path), f'{full_path} does not exist'
            assert os.path.isdir(full_path), f'{full_path} is not a directory'
            assert confirm_directory_download_correct(temp_dir, rel_path, contents), f'The content of local dir "{full_path}" does not match {s3_uri}/{rel_path}'
    else:
        reason = populated_s3_checkpoints['reason']
        warning(f'Skipping {current_test_name()}() because {reason}')
        assert True


def test_download_all_sync(populated_s3_checkpoints):
    s3_uri = populated_s3_checkpoints['uri']
    if s3_uri:
        contents = populated_s3_checkpoints['contents']

        # print(f'trying download_one_sync on rel_path "{rel_path}"')
        with tempfile.TemporaryDirectory() as temp_dir:
            synchronizer = RemoteCheckpointSynchronizer.from_uri(s3_uri, local_output_dir = temp_dir)
            assert synchronizer
            synchronizer.download_all_sync()
            assert confirm_directory_download_correct(temp_dir,'',contents), f'The content of local dir "{temp_dir}" does not match {s3_uri}'
    else:
        reason = populated_s3_checkpoints['reason']
        warning(f'Skipping {current_test_name()}() because {reason}')
        assert True


from .utils.temp_s3_object import TempS3Object
from .utils.s3_dict import S3Dict


checkpoints_content:dict[str,str] = {
    'checkpoint-1000/c1f1': 'local content of checkpoint-1000/c1f1',
    'checkpoint-1000/c1f2': 'local content of checkpoint-1000/c1f2',
    'checkpoint-1000/c1f3': 'local content of checkpoint-1000/c1f3',
    'checkpoint-1000/c1d1/c1d1f1': 'local content of checkpoint-1000/c1d1/c1d1f1',
    'checkpoint-1000/c1d1/c1d1f2': 'local content of checkpoint-1000/c1d1/c1d1f2',
    'checkpoint-1000/c1d1/c1d2f1': 'local content of checkpoint-1000/c1d1/c1d2f1',
    'checkpoint-1000/c1d1/c1d2f2': 'local content of checkpoint-1000/c1d1/c1d2f2',
    'checkpoint-2000/c2f1': 'local content of checkpoint-2000/c2f1',
    'checkpoint-2000/c2f2': 'content of checkpoint-2000/c2f2',
    'checkpoint-2000/c2f3': 'content of checkpoint-2000/c2f3',
    'checkpoint-2000/c2d1/c1d1f1': 'content of checkpoint-2000/c2d1/c1d1f1',
    'checkpoint-2000/c2d1/c1d1f2': 'content of checkpoint-2000/c2d1/c1d1f2',
    'checkpoint-2000/c2d1/c1d2f1': 'content of checkpoint-2000/c2d1/c1d2f1',
    'checkpoint-2000/c2d1/c1d2f2': 'content of checkpoint-2000/c2d1/c1d2f2',
    'checkpoint_registry.json': '{"1000": {"checkpoint_name": "checkpoint-1000", "global_step": 1000, "segment_number": 1}, "2000": {"checkpoint_name": "checkpoint-2000", "global_step": 2000, "segment_number": 2} }'
}

def test_upload_all_aync(writable_s3_uri):
    if writable_s3_uri:
        bucket, prefix = TempS3Object.parse_s3_uri(writable_s3_uri)
        if prefix: # suppress pyright warnings about unused var
            pass
        with tempfile.TemporaryDirectory() as temp_local_dir:
            copy_dict_to_fs(checkpoints_content, temp_local_dir)
            with TempS3Object(writable_s3_uri).TemporaryDirectory(delete=True) as temp_s3_dir:
                dest_s3_uri = f's3://{bucket}/{temp_s3_dir}'
                synchronizer = RemoteCheckpointSynchronizer.from_uri(dest_s3_uri, local_output_dir = temp_local_dir)
                task = synchronizer.upload_all_async()
                task.wait_for_it()
                assert not task.error
                s3_dict = S3Dict(bucket)
                for rel_path, file_content in checkpoints_content.items():
                    full_key = f'{temp_s3_dir}/{rel_path}'
                    assert s3_dict[full_key] == file_content

    else:
        reason = 'No writable S3 URI defined'
        warning(f'Skipping {current_test_name()}() because {reason}')




from .continuing_trainer import QLoRAContinuingTrainer, logger
from .checkpoint_manager import CheckpointManager
from .checkpoint_registry import CheckpointInfo, CheckpointRegistry, RemoteCheckpointSynchronizer


# silence the PyRight unused symbol warnings
if QLoRAContinuingTrainer or logger or CheckpointManager or CheckpointInfo or CheckpointRegistry or RemoteCheckpointSynchronizer:
    pass



#!/usr/bin/env python
# coding: utf-8


import re
from typing import Union, cast
import torch
from .checkpoint_registry import CheckpointRegistry, CheckpointInfo, RemoteCheckpointSynchronizer
from .checkpoint_manager import CheckpointManager
from s3datasets import S3TextDataset
from datasets import Dataset, IterableDataset
from tokengenerators import TextDS2TokensGenerator

from peft.mapping import  get_peft_model
from peft.peft_model import PeftModel
from peft import LoraConfig # type: ignore
from peft.utils.other import prepare_model_for_kbit_training
from abc import ABC, abstractmethod

from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    PreTrainedTokenizerBase,
    PreTrainedModel,
    BitsAndBytesConfig,
    TrainingArguments,
    Trainer
)

def set_up_logging():
    import logging
    import sys
    logger = logging.getLogger('standard')
    logger.setLevel(logging.DEBUG)

    info_fh = logging.FileHandler('continuing_trainer.info.log')
    info_fh.setLevel(logging.INFO)
    info_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    info_fh.setFormatter(info_formatter)

    debug_fh=logging.FileHandler('continuing_trainer.debug.log')
    debug_fh.setLevel(logging.DEBUG)
    debug_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s in %(funcName)s at %(filename)s:%(lineno)s')
    debug_fh.setFormatter(debug_formatter)

    err_handler = logging.StreamHandler(sys.stderr)
    err_handler.setLevel(logging.ERROR)


    out_handler = logging.StreamHandler(sys.stdout)
    out_handler.setLevel(logging.INFO)


    logger.addHandler(info_fh)
    logger.addHandler(debug_fh)
    logger.addHandler(err_handler)
    logger.addHandler(out_handler)
    return logger

logger = set_up_logging()




def torchify_collator(batch):
    # Initialize empty lists to hold each attribute
    labels = []
    attention_mask = []

    # Iterate through each item in the batch, grouping by column
    batched_columns = {}
    for item in batch:
        for column_name,v in item.items():
           if not column_name in batched_columns:
               batched_columns[column_name] = []
           batched_columns[column_name].append(v)

        labels.append(item['labels'])
        attention_mask.append(item['attention_mask'])

    # Convert lists to tensors
    collated = {}
    for column_name,column_data in batched_columns.items():
        if isinstance(column_data[0], list):
            collated[column_name] = torch.tensor(column_data,dtype=torch.long)
        else:
            collated[column_name] = column_data
    return collated


class BaseContinuingTrainer(ABC):
    """
    An abstract base class for trainers that support the continuation of training LoRA-enhanced Huggingface
    transformer models on text documents stored in an S3 bucket, designed for scenarios where training sessions
    may be interrupted such as on vast.ai or salad.com instances or anywhere where intermittent harware
    failure is reasonable likely to occur (everywhere?).

    This class aims to minimize the time, network bandwidth, and storage needed to restart training by efficiently
    resuming from the point of interruption, leveraging the `s3datasets` library for handling dataset lazy
    downloading and tokenizing/chunking.

    In practice, this base class is not used directly but instead, a subclass such as QLoRAContinuingTrainer
    is instantiated.

    Additional subclasses can be implemented to enable different quantization and PEFT techniques.  e.g. LoftQ

    The following abstract methods defined in this class must be overridden in subclasses to specify model
    loading behaviors:
    - `load_checkpoint_model(checkpoint_path, base_model_id)`: Load model from a saved checkpoint.
    - `load_starting_model(base_model_id)`: Prepare a starting model based on `base_model_id`.

    Args:
        base_model_id (str): Identifier for the base model to which LoRA will be attached.
        dataset_bucket_name (str): Name of the S3 bucket containing the training text documents.
        output_dir (str, optional): Directory where checkpoints will be saved. Defaults to "/root/outputs".
        dataset_id (Union[str, None], optional): Key to an S3 object containing JSON array of keys for training
            text objects.
        dataset_series (Union[str, None], optional): Pattern for S3 keys of JSON objects specifying keys of
            objects for training.
            Must include "{segment_number}" if specified. Defaults to None.
            Exactly one of `dataset_id` or `dataset_series` must be specified. Defaults to None.
        test_dataset_id (Union[str, None], optional): Key to an S3 object with JSON array of keys for evaluation
            text objects. Defaults to None.
        steps_per_round (Union[int, None], optional): Maximum number of training steps per call to train(). Defaults
            to None.
        max_seq_length (int, optional): Maximum token count per training example. Defaults to 2048.
        max_steps (Union[int, None]): Explicit maximum training steps.  Future implementations will
            estimate this if it is not supplied by the caller but for now it is required.
        save_steps (int, optional): Interval of training steps after which a checkpoint is auto-saved.
            Defaults to 1000.

    Raises:
        ValueError: If both or neither `dataset_id` and `dataset_series` are specified, or if `dataset_series`
        does not contain "{segment_number}".
    """

    # members initialized from constructor params
    base_model_id: str # the id of the base model which will have a LoRA attached to it
    dataset_bucket_name: str # The name of an S3 bucket containing text documents on which to train
    output_dir: str # The filesystem directory where checkpoints should be saved
    dataset_id: Union[str,None]  # A key to an S3 object containing a JSON array of keys of the text objects on which to train
    dataset_series: Union[str,None] # A pattern for S3 keys of json objects specifying keys of the specific objects to use in training
    test_dataset_id: Union[str,None] # A key to an S3 object containing a JSON array of keys of the text objects to use an an evaluation dataset
    steps_per_round: Union[int,None] # Maximum number of training steps to execute in a single call to train().  This is mostly useful for testing the process of interrupting and then continuing training.  Once that's been tested, just make this a very large number.
    max_seq_length: int # maximum number of tokens in a single training example (GPU memory constrains this )
    save_steps: int # Automatically make a checkpoint when this many training steps are completed
    explicit_max_steps: Union[int,None] # If the caller specified max_steps during construction, the value is stored here

    # internally initialized members
    checkpoint_registry: CheckpointRegistry  # Manages the registration and retrieval of training checkpoints.
    tokenizer: PreTrainedTokenizerBase # The tokenizer used to convert text to tokens (derived from base_model_id)
    base_model: PreTrainedModel # The loaded base model
    model: PeftModel # The model to be trained, endowed with trainable LoRA modules
    starting_step: Union[int,None]   # The step number at which the current round of training starts
    starting_checkpoint_info: Union[CheckpointInfo,None] # Full details on the starting checkpoint if continuing
    text_dataset: Dataset # A datset of text documents produced by S3TextDataset
    train_tokens_generator: TextDS2TokensGenerator # A generator of tokenized examles drawing from text_dataset that is used as the basis of the train_inputs iterable dataset
    train_inputs: IterableDataset # A dataset of training examples ready to feed to a transformers.Trainer
    test_inputs: Union[Dataset, IterableDataset,None] # A dataset of validation examples ready to feed to a transformers.Trainer
    checkpoint_manager: CheckpointManager # A TrainerCallback-based manager to create checkpoints and register them and the corresponding generator cursors

    def __init__(self,
                 base_model_id: str,
                 dataset_bucket_name: str,
                 output_dir:str = "/root/outputs",
                 checkpoint_sync_uri: Union[str,None] = None, # typically s3://{bucketname}/{prefix}
                 dataset_id: Union[str,None] = None,
                 dataset_series: Union[str,None] = None,
                 test_dataset_id: Union[str,None] = None,
                 steps_per_round: Union[int,None] = None,
                 max_seq_length: int = 2048,
                 max_steps: Union[int,None] = None,
                 save_steps: int = 1000

                 ):
        if (dataset_id and dataset_series) or (dataset_id == None and dataset_series == None):
            raise ValueError("exactly one of dataset_id or dataset_series must be specified")
        if dataset_series and not '{segment_number}' in dataset_series:
            raise ValueError('dataset_series if specified must contain the substring "{segment_number}"')
        self.base_model_id = base_model_id
        self.dataset_bucket_name = dataset_bucket_name
        self.output_dir= output_dir
        self.dataset_id = dataset_id
        self.test_dataset_id = test_dataset_id
        self.dataset_series = dataset_series
        self.steps_per_round = steps_per_round
        self.max_seq_length = max_seq_length
        self.explicit_max_steps = max_steps
        self.save_steps = save_steps

        checkpoint_synchronizer: Union[RemoteCheckpointSynchronizer,None] = None
        if checkpoint_sync_uri:
            checkpoint_synchronizer = RemoteCheckpointSynchronizer.from_uri(uri=checkpoint_sync_uri, local_output_dir=output_dir)

        self.checkpoint_registry = CheckpointRegistry(output_dir=output_dir,
                                                      remote_synchronizer=checkpoint_synchronizer)
        self.starting_checkpoint_info = None

        self.prepare()

    @abstractmethod
    def load_checkpoint_model(self, checkpoint_path:str, base_model_id: str) -> PeftModel:
        """
            This method must be overridden by subclasses.
            Its purpose is to load the model to be trained from a previously saved checkpoint saved
            in the directory specified by the checkpoint_path parameter.

            e.g.  Load a QLoRA model from a checkpoint using the checkpoint_path and base_model_id
        """
        if checkpoint_path or base_model_id: # prevent PyRight warnings about unused params
            pass

    @abstractmethod
    def load_starting_model(self, base_model_id:str) -> PeftModel:
        """
            This method must be overridden by subclasses.
            Its purpose is to load/prepare the model to be trained from a base model.

            e.g.  Load and quantize the base model and attach a no-op LoRA to it for QLoRA training
        """
        if base_model_id: # prevent PyRight warnings about unused params
            pass

    def prepare(self):
        self._cuda_check()
        self._get_continuation_state()
        self._load_model()
        self._prepare_datasets()
        self._prepare_trainer()

    def train(self):
        # pre-requisites
        assert self.trainer
        assert not self.starting_step == None

        logger.info(f'generator cursor before training is segment {self.dataset_segment_number()},  {self.train_tokens_generator.get_cursor().to_dict()}')
        try:
            if self.starting_checkpoint_info:
                logger.info(f'Resuming training from checkpoint {self.starting_checkpoint_info.path()}')
                self.trainer.train(resume_from_checkpoint=self.starting_checkpoint_info.path(),
                              without_checkpoint_model=True) # We already loaded the model we want, don't screw it up
            else:
                logger.info(f'Starting training from the beginning.')
                self.trainer.train()
        except ValueError as e:
            # check if the message matches the regex '^Batch does not contain any data'
            if re.match('^Batch does not contain any data', str(e)):
                logger.info(f'ran out of data and the dataset_generator cursor is {self.train_tokens_generator.get_cursor().to_dict()}')
                # Calling this because the stupid exception thrown above prevented it from being invoked
                # It needs to be called to ensure the last state of the model is saved.
                self.checkpoint_manager.on_train_end(self.trainer.args, self.trainer.state, self.trainer.control)

        logger.info(f'generator cursor after training is segment {self.dataset_segment_number()},  {self.train_tokens_generator.get_cursor().to_dict()}')
        logger.info(f'Trained for {self.checkpoint_manager.steps_seen} steps')
        logger.debug(f'post training trainer state {self.trainer.state.__repr__()}')
        # Ensure that any ongoing checkpoint uploads complete before returning
        self.checkpoint_registry.finish_up()

    def reached_max_steps(self) -> bool:
        return self.checkpoint_manager and self.checkpoint_manager.max_steps_passed

    def dataset_segment_number(self):
        if 'checkpoint_manager' in self.__dict__ and self.checkpoint_manager:
            return self.checkpoint_manager.current_dataset_segment_number
        elif self.starting_checkpoint_info and self.starting_checkpoint_info.segment_number:
            return self.starting_checkpoint_info.segment_number
        else:
            return 1

    def _get_continuation_state(self) -> int:
        self.starting_step = self.checkpoint_registry.latest_step()
        logger.debug(f"starting_step is initially {self.starting_step}")
        if self.starting_step:
           self.starting_checkpoint_info = self.checkpoint_registry.get_checkpoint_for_step(self.starting_step)
           if not self.starting_checkpoint_info:
               raise RuntimeError(f"starting_step is defined as {self.starting_step} but could not retrieve checkpoint_info for it")
           # check if self.previous_checkpoint_path exists and raise an error if it does not
           if not self.starting_checkpoint_info.exists():
               raise RuntimeError(f"starting checkpoint path is defined as {self.starting_checkpoint_info.path()} but it does not exist")
           logger.info(f"resuming training from checkpoint {self.starting_checkpoint_info.__dict__.__repr__()}")
        else:
           self.previous_checkpoint_path = None
           self.starting_step = 0
        return self.starting_step

    def _load_model(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self.base_model_id)
        self.tokenizer.pad_token = self.tokenizer.eos_token

        if self.starting_checkpoint_info:
            checkpoint_path=self.starting_checkpoint_info.path()
            logger.info(f"Loading previous checkpoint model at {checkpoint_path}")
            self.model = self.load_checkpoint_model(checkpoint_path, self.base_model_id)
        else:
            logger.info(f"Starting with untrained LoRA on {self.base_model_id}")
            self.model = self.load_starting_model(self.base_model_id)
        assert self.model
        logger.info(f"LoRA model stats: {self.trainable_parameters_description(self.model)}")
        self.model.config.use_cache = False  # silence the warnings. Please re-enable for inference!
        return self.model

    def _prepare_datasets(self) -> tuple[IterableDataset, Union[Dataset,None]]:
        assert self.tokenizer

        train_dataset_id = self._current_dataset_id()
        logger.debug(f"Preparing train dataset from {train_dataset_id}")
        self.text_dataset = S3TextDataset(self.dataset_bucket_name, dataset_id=train_dataset_id).to_full_dataset()
        logger.info(f'text_dataset contains {len(self.text_dataset)} documents')
        # TODO:  maybe make these constructor params?
        min_stride=64
        max_waste=64
        self.train_tokens_generator = TextDS2TokensGenerator(self.text_dataset, self.tokenizer, chunk_len = self.max_seq_length, min_stride = min_stride, max_waste = max_waste)
        features = self.train_tokens_generator.features()
        if self.test_dataset_id:
            test_text_dataset = S3TextDataset(self.dataset_bucket_name, dataset_id=self.test_dataset_id).to_full_dataset()
            test_tokens_generator = TextDS2TokensGenerator(test_text_dataset, self.tokenizer, chunk_len = self.max_seq_length, min_stride = min_stride, max_waste = max_waste)
            test_inputs_ds: Dataset = cast(Dataset, Dataset.from_generator(test_tokens_generator))
            self.test_inputs = test_inputs_ds
        else:
            self.test_inputs = None
        self.train_inputs = IterableDataset.from_generator(self.train_tokens_generator,features=features)
        return (self.train_inputs, self.test_inputs)

    def estimated_max_steps(self):
        raise RuntimeError("Have not yet implemented estimated_max_steps(), so specify max_steps explicitly at construction.")

    def _prepare_trainer(self) -> Trainer:
        # pre-requisites
        assert self.model
        assert self.train_inputs
        assert self.train_tokens_generator

        data_collator = torchify_collator
        num_train_epochs = 1
        gradient_accumulation_steps = 4
        #max_steps = estimated_number_of_training_examples // gradient_accumulation_steps
        max_steps = self.explicit_max_steps if self.explicit_max_steps else self.estimated_max_steps()

        warmup_steps = 8 if self.dataset_segment_number() in [1,None] else 0
        starting_learning_rate = 2e-4

        training_args = TrainingArguments(
                ignore_data_skip = True, # when resuming, do not try to skip past previously trained examples.  The train_tokens_generator.set_cursor() call takes care of that.
                remove_unused_columns = True,
                per_device_train_batch_size=1,
                per_device_eval_batch_size=1,
                gradient_accumulation_steps=gradient_accumulation_steps,
                warmup_steps=warmup_steps,
                learning_rate=starting_learning_rate,
                fp16=True,
                logging_steps=1,
                output_dir=self.output_dir,
                optim="paged_adamw_8bit",
                num_train_epochs = num_train_epochs,
                save_steps=self.save_steps,      # Checkpoint regularly every this many steps
                max_steps=max_steps,
                evaluation_strategy="no",  # Do not bother with evaluation
                include_num_input_tokens_seen = True,
                #eval_steps = eval_steps,  # Uncomment and set this if you choose evaluation_strategy="steps"
            )


        self.trainer = Trainer(
                        model=self.model,
                        train_dataset=self.train_inputs, # type: ignore (IterableDataset is apparently not envisioned here)
                        eval_dataset=self.test_inputs, # type: ignore (IterableDataset is apparently not envisioned here)
                        args=training_args,
                        data_collator=data_collator
        )
        # We want to be able to restart the trainer at will so instead of putting max_steps
        # in the trainer's arguments, we put it in a callback that breaks early on an infinite steps
        # overall training run

        # it will add itself as a callback to the trainer in its constructor
        self.checkpoint_manager = CheckpointManager(self.trainer,
                                                  self.checkpoint_registry,
                                                  self.train_tokens_generator,
                                                  starting_checkpoint_info=self.starting_checkpoint_info,
                                                  stop_after_steps = self.steps_per_round,
                                                  logger=logger)
        return self.trainer


    def _cuda_check(self) -> None:
        use_cuda = torch.cuda.is_available()
        if use_cuda:
            logger.debug(f'__CUDNN VERSION: {torch.backends.cudnn.version()}')
            logger.debug(f'__Number CUDA Devices: {torch.cuda.device_count()}')
            logger.debug(f'__CUDA Device Name: {torch.cuda.get_device_name(0)}')
            logger.debug(f'__CUDA Device Total Memory [GB]: {torch.cuda.get_device_properties(0).total_memory/1e9}')

    def _current_dataset_id(self) -> str:
        if self.dataset_id:
            return self.dataset_id
        else:
            assert self.dataset_series
            dataset_id = self.dataset_series.replace("{segment_number}",str(self.dataset_segment_number()))
            return dataset_id

    @staticmethod
    def trainable_parameters_description(model) -> str:
        """
        Reports the number of trainable parameters in a model.
        """
        trainable_params = 0
        all_param = 0
        for _, param in model.named_parameters():
            all_param += param.numel()
            if param.requires_grad:
                # print(_)
                trainable_params += param.numel()
        return f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}"




class QLoRAContinuingTrainer(BaseContinuingTrainer):
    """
    Implements the continuation of training for NF4-quantized LoRA-enhanced Hugging Face Transformer models.

    This class extends `BaseContinuingTrainer`, implementing model loading methods specific to using QLoRA
    which quantizes a base model with NF4 and then applies a trainable Low Rank Adapter to it.

    For a detailed explanation of functionality and constructor parameters, see the base class `BaseContinuingTrainer`.

    Note that the standard QLoRA model instantiation which this class uses requires the downloading of a high
    precision model which is then dynamically quantized to NF4 during loading onto the GPU.
    """

    def _load_base_model(self, base_model_id: str) -> PreTrainedModel:
        logger.debug(f"loading base model {self.base_model_id}")
        bnb_config = BitsAndBytesConfig(
            load_in_4bit=True,
            load_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
        base_model = AutoModelForCausalLM.from_pretrained(
            base_model_id,
            device_map="auto",
            trust_remote_code=True,
            quantization_config=bnb_config
        )
        base_model.gradient_checkpointing_enable()         # significantly the reduce memory footprint of the activations cache during training
        base_model = prepare_model_for_kbit_training(base_model)
        return base_model

    def announce_model_config(self, action:str, model: PeftModel) -> None:
        active_adapter = model.active_adapter
        # sometimes active_adaapter is a string, sometimes a function, ???
        if not isinstance(active_adapter, str):
            active_adapter = active_adapter()
        logger.info(f"{action} Q-LoRA model. active adapter: {active_adapter} ")

    def load_checkpoint_model(self, checkpoint_path: str, base_model_id: str) -> PeftModel:
        if not self.__dict__.get('base_model',None):
            self.base_model = self._load_base_model(base_model_id)
        assert self.base_model
        model = PeftModel.from_pretrained(model=self.base_model,
                                          model_id=checkpoint_path,
                                          adapter_name="default",
                                          is_trainable=True)
        self.announce_model_config('Loaded', model)
        return model

    def build_adapter_config(self) -> LoraConfig:
        mistral_7b_target_modules =   [
                        'q_proj',
                        'k_proj',
                        'v_proj',
                        'o_proj',
                        'gate_proj',
                        'up_proj',
                        'down_proj',
                        'lm_head', # unsloth does not use this in their benchmark, why?
        ]

        adapter_config = LoraConfig(
            r=16,
            lora_alpha=32,
            target_modules=mistral_7b_target_modules,
            lora_dropout=0,  # Initially 0.05, changed to 0 because of unsloth
            bias="none",
            task_type="CAUSAL_LM",
        )
        return adapter_config

    def load_starting_model(self, base_model_id: str) -> PeftModel:
        if not self.__dict__.get('base_model',None):
            self.base_model = self._load_base_model(base_model_id)
        assert self.base_model
        model = cast(PeftModel,get_peft_model(self.base_model,self.build_adapter_config()))
        self.announce_model_config('Constructed', model)
        return model




import tempfile
import os
import sys

path_to_source_modules=os.path.abspath(os.path.join(os.path.dirname(__file__), '../src'))
sys.path.insert(0, path_to_source_modules)
import continuing_education
from continuing_education import CheckpointRegistry, RemoteCheckpointSynchronizer
continuing_education.checkpoint_registry.verbose = False # Make this True for debugging

from .utils.misc import  copy_dict_to_s3, copy_dict_to_fs
from .utils.directory_dict import DirectoryDict
from .utils.s3_dict import S3Dict
from .utils.temp_s3_object import TempS3Object
from .utils.persistent_temporary_directory import PersistentTemporaryDirectory
if PersistentTemporaryDirectory: # Supress pyright unused symbol warnings
    pass

checkpoints_content:dict[str,str] = {
    'checkpoint-1000/c1f1': 'local content of checkpoint-1000/c1f1',
    'checkpoint-1000/c1f2': 'local content of checkpoint-1000/c1f2',
    'checkpoint-1000/c1f3': 'local content of checkpoint-1000/c1f3',
    'checkpoint-1000/c1d1/c1d1f1': 'local content of checkpoint-1000/c1d1/c1d1f1',
    'checkpoint-1000/c1d1/c1d1f2': 'local content of checkpoint-1000/c1d1/c1d1f2',
    'checkpoint-1000/c1d1/c1d2f1': 'local content of checkpoint-1000/c1d1/c1d2f1',
    'checkpoint-1000/c1d1/c1d2f2': 'local content of checkpoint-1000/c1d1/c1d2f2',
    'checkpoint-2000/c2f1': 'local content of checkpoint-2000/c2f1',
    'checkpoint-2000/c2f2': 'content of checkpoint-2000/c2f2',
    'checkpoint-2000/c2f3': 'content of checkpoint-2000/c2f3',
    'checkpoint-2000/c2d1/c1d1f1': 'content of checkpoint-2000/c2d1/c1d1f1',
    'checkpoint-2000/c2d1/c1d1f2': 'content of checkpoint-2000/c2d1/c1d1f2',
    'checkpoint-2000/c2d1/c1d2f1': 'content of checkpoint-2000/c2d1/c1d2f1',
    'checkpoint-2000/c2d1/c1d2f2': 'content of checkpoint-2000/c2d1/c1d2f2',
    'checkpoint_registry.json': '{"1000": {"checkpoint_name": "checkpoint-1000", "global_step": 1000, "segment_number": 1}, "2000": {"checkpoint_name": "checkpoint-2000", "global_step": 2000, "segment_number": 2}}'
}

def test_construction_with_remote_synchronizer(writable_s3_uri):
    bucket, prefix = TempS3Object.parse_s3_uri(writable_s3_uri)
    if prefix: # suppress pyright warnings about unused var
        pass
    with TempS3Object(writable_s3_uri).TemporaryDirectory(delete=True) as temp_s3_dir:
        temp_s3_uri = f's3://{bucket}/{temp_s3_dir}'
        copy_dict_to_s3(checkpoints_content, temp_s3_uri)
        with tempfile.TemporaryDirectory() as temp_local_dir:
        # with PersistentTemporaryDirectory() as temp_local_dir:
            synchronizer_s3_uri = f's3://{bucket}/{temp_s3_dir}'

            synchronizer = RemoteCheckpointSynchronizer.from_uri(synchronizer_s3_uri, local_output_dir = temp_local_dir)
            registry = CheckpointRegistry(output_dir = temp_local_dir, remote_synchronizer = synchronizer)
            assert registry._registry[1000].checkpoint_name == 'checkpoint-1000'
            assert registry._registry[2000].checkpoint_name == 'checkpoint-2000'
            assert registry.latest_step() == 2000
            checkpoint2000 = registry.get_checkpoint_for_step(2000)
            assert checkpoint2000
            checkpoint_name = checkpoint2000.checkpoint_name 
            checkpoint_full_path = os.path.join(temp_local_dir, checkpoint_name) 
            assert os.path.exists(checkpoint_full_path)
            assert os.path.isdir(checkpoint_full_path)
            dir_dict = DirectoryDict(temp_local_dir)
            for rel_path in checkpoints_content.keys():
                full_path = os.path.join(temp_local_dir,rel_path)
                # if it begins with checkpoint_name
                if rel_path.startswith(checkpoint_name+'/') or rel_path == 'checkpoint_registry.json':
                    assert os.path.exists(full_path)
                    expected_content = checkpoints_content[rel_path]
                    assert dir_dict[rel_path] == expected_content
                else:
                    assert not os.path.exists(full_path)

def test_construction_without_remote_synchronizer_or_data():
    with tempfile.TemporaryDirectory() as temp_local_dir:
        registry = CheckpointRegistry(output_dir = temp_local_dir)
        assert registry
        assert registry.latest_step() == None

def test_construction_without_remote_synchronizer_but_local_data():
    with tempfile.TemporaryDirectory() as temp_local_dir:
        copy_dict_to_fs(checkpoints_content, temp_local_dir)
        registry = CheckpointRegistry(output_dir = temp_local_dir)
        assert registry
        latest_step = registry.latest_step()
        assert latest_step == 2000
        checkpoint_2000 = registry.get_checkpoint_for_step(latest_step)
        assert checkpoint_2000
        dd = DirectoryDict(checkpoint_2000.path())
        for checkpoint_rel_path in dd.keys():
            output_dir_rel_path = f'{checkpoint_2000.checkpoint_name}/{checkpoint_rel_path}'
            assert dd[checkpoint_rel_path] == checkpoints_content[output_dir_rel_path]

def test_uploads_of_new_checkpoints(writable_s3_uri):
    synchronizer_s3_uri = writable_s3_uri + '.extra_temp'
    with tempfile.TemporaryDirectory() as temp_local_dir:
        copy_dict_to_fs(checkpoints_content, temp_local_dir)
        synchronizer = RemoteCheckpointSynchronizer.from_uri(synchronizer_s3_uri, local_output_dir = temp_local_dir)
        registry = CheckpointRegistry(output_dir = temp_local_dir, remote_synchronizer = synchronizer)
        # initiate an async upload
        registry.save()
        assert registry.upload_in_progress
        registry.finish_up()
        assert not registry.upload_in_progress
        bucket, prefix = TempS3Object.parse_s3_uri(synchronizer_s3_uri)
        s3d = S3Dict(bucket)
        for s3_path in s3d.keys():
            if s3_path.startswith(prefix+'/'):
                checkpoint_rel_path = s3_path[len(prefix)+1:]
                assert s3d[s3_path] == checkpoints_content[checkpoint_rel_path]





import os
import sys
if sys: # suppress pyright warnings about unused sys when all diagnostic prints are commented out
    pass

import re
import json
from typing import Callable, Dict, Union, Any, cast, Type
from dataclasses import dataclass
from abc import ABC, abstractmethod
import threading
import subprocess
import queue
from queue import Queue
import boto3
from botocore.exceptions import ClientError
from datetime import timezone

REGISTRY_FILENAME='checkpoint_registry.json'

verbose = False

def debug_print(message:str, file=sys.stdout):
    if verbose:
        print(message, file=file)

@dataclass
class CheckpointInfo:
    checkpoint_name: str
    global_step: int
    segment_number: Union[int,None]
    output_dir: str

    def mini_dict(self) -> Dict[str,Any]:
        return {k: v for k,v in self.__dict__.items() if k != 'output_dir'}

    def path(self) -> str:
        return os.path.join(self.output_dir,self.checkpoint_name)

    def exists(self) -> bool:
        return os.path.exists(self.path())

    @classmethod
    def from_mini_dict(cls, output_dir, mini_dict: Dict[str,Any]) -> 'CheckpointInfo':
        instance = cls(output_dir = output_dir, **mini_dict)
        return instance


import time

class RemoteCheckpointSynchronizer(ABC):
    local_output_dir: str
    remote_uri: str
    draining_the_queue: bool

    _handler_registry: list[Type['RemoteCheckpointSynchronizer']] = []
    _worker_thread: threading.Thread

    def __init__(self, local_output_dir: str, remote_uri: str = ""):

        assert self.__class__.can_handle_uri(remote_uri)

        self.local_output_dir = local_output_dir
        self.remote_uri = remote_uri

        self.draining_the_queue = False
        self._task_queue = RemoteCheckpointSynchronizer.TaskQueue()
        self._worker_thread = threading.Thread(target=self._work_the_queue)
        self._worker_thread.daemon = True  # Ensure thread exits when main program exits
        self._worker_thread.start()


    def _work_the_queue(self):
        all_done = False
        while not all_done:
            task: Union[RemoteCheckpointSynchronizer.Task,None] = None
            try:
                task = self._task_queue.get(block=True, timeout=1.0)
            except queue.Empty:
                if self.draining_the_queue:
                    all_done = True
            if task:
                try:
                    task.execute()
                    task.done = True
                finally:
                    self._task_queue.task_done()  # Mark the task as done

    def finish_up(self):
        self.draining_the_queue = True
        self._worker_thread.join()

    @classmethod
    def register_thyself(cls: Type['RemoteCheckpointSynchronizer']) -> None:
        cls._handler_registry.append(cls)

    @classmethod
    @abstractmethod
    def can_handle_uri(cls, uri: str) -> bool:
        """ Subclasses must override this to return whether or not they can handle the uri in question """
        return False

    @classmethod
    def handler_for_uri(cls,uri) -> Type['RemoteCheckpointSynchronizer']:
        for handler in cls._handler_registry:
            if handler.can_handle_uri(uri):
                return handler
        raise RuntimeError(f'No RemoteCheckpointSynchronizer exists that can handle the uri "{uri}"')

    @classmethod
    def from_uri(cls, uri:str, local_output_dir:str) -> 'RemoteCheckpointSynchronizer':
        klass:Type['RemoteCheckpointSynchronizer'] = cls.handler_for_uri(uri)
        assert klass
        return  klass(local_output_dir = local_output_dir, remote_uri=uri)

    @abstractmethod
    def _upload_all_sync(self):
        """ Subclasses must override this to implement a synchronous upload of everything in the local_output_dir """
        pass

    @abstractmethod
    def download_all_sync(self) -> bool:
        """
            Subclasses must overrride this.
            Downloads all files and subdirectories of the checkpoints directory from the remote to local.
            Returns true if the the local directory exists after the operation.
        """
        pass

    @abstractmethod
    def download_one_sync(self,relative_path: str) -> Union[str,None]:
        """ downloads one file or directory under the local_dir and returns either the full path to it or None if it failed to download """
        pass

    def upload_all_async(self) -> 'RemoteCheckpointSynchronizer.Task':
        # Add the sync operation to the queue
        task = RemoteCheckpointSynchronizer.Task(self._upload_all_sync,{})
        self._task_queue.put(task)
        return task

    class Task:
        func: Callable
        kwargs: dict
        done: bool
        result: Any
        error: Union[BaseException, None]

        def __init__(self, func:Callable, kwargs:dict) -> None:
            self.func = func
            self.kwargs = kwargs
            self.done=False
            self.error = None
            self.result = None

        def execute(self) -> None:
            try:
                self.result = self.func(**self.kwargs)
            except Exception as e:
                self.error = e
            finally:
                self.done=True

        def wait_for_it(self, poll_interval = 0.1) -> None:
            while not self.done:
                time.sleep(poll_interval)

    class TaskQueue(Queue):
        def put(self, item: 'RemoteCheckpointSynchronizer.Task', *args, **kwargs) -> None:
            return super().put(item, *args, **kwargs)

        def get(self, *args, **kwargs) -> 'RemoteCheckpointSynchronizer.Task':
            return cast('RemoteCheckpointSynchronizer.Task',super().get(*args,**kwargs))




class S3RemoteCheckpointSynchronizer(RemoteCheckpointSynchronizer):

    def _upload_all_sync(self):
        cmd = f'aws s3 sync "{self.local_output_dir}" "{self.remote_uri}"'
        if not verbose:
            cmd += ' > /dev/null 2>&1'
        subprocess.run(cmd, shell=True, check=True)

    def download_one_sync(self, relative_path: str) -> Union[str,None]:
        full_local_path = os.path.join(self.local_output_dir,relative_path)
        full_uri = f"{self.remote_uri}/{relative_path}"
        # first check if it is an ordinary file and if so, follow sync logic on it.
        # aws s3 sync command only works on directories so I had to implement this:
        if not S3RemoteCheckpointSynchronizer.sync_s3_uri_to_local_file(full_uri,full_local_path):
            debug_print(f'The remote {full_uri}  does not exist as an object, checking to see if it is a dir of objects')
            if S3RemoteCheckpointSynchronizer._remote_dir_exists(full_uri):
                debug_print(f'{full_uri} appears to be a dir of objects.  Using s3 sync')
                # use s3 sync in case the source is a directory
                cmd = f'aws s3 sync "{full_uri}" "{full_local_path}"'
                if not verbose:
                    cmd += ' > /dev/null 2>&1'
                debug_print(f'about to run command "{cmd}"', file=sys.stderr)
                subprocess.run(cmd, shell=True, check=True)
            else:
                debug_print(f'{full_uri} does not match anything, file nor dir')
        return full_local_path if (os.path.exists(full_local_path)) else None

    def download_all_sync(self) -> bool:
        cmd = f'aws s3 sync "{self.remote_uri}" "{self.local_output_dir}"'
        if not verbose:
            cmd += ' > /dev/null 2>&1'
        subprocess.run(cmd, shell=True, check=True)
        return os.path.exists(self.local_output_dir)

    @classmethod
    def can_handle_uri(cls, uri: str) -> bool:
        # does it match s3: ?
        return bool(re.match("^s3://",uri))

    @staticmethod
    def parse_s3_uri(s3_uri:str) -> tuple[str,str]:
        uri_pattern = r'^s3://([^/]+)/(.+)$'
        match = re.match(uri_pattern, s3_uri)
        if match:
            # Extract bucket name and key from the matched groups
            bucket,key = match.groups()
            return (bucket,key)
        else:
            raise ValueError(f"Invalid S3 URI format: {s3_uri}")

    @staticmethod
    def sync_s3_uri_to_local_file(s3_uri: str, local_file_path: str) -> bool:
        # Parse the S3 URI
        bucket_name, key = S3RemoteCheckpointSynchronizer.parse_s3_uri(s3_uri)

        # Create an S3 client
        s3_client = boto3.client('s3')

        try:
            # Get object metadata
            metadata = s3_client.head_object(Bucket=bucket_name, Key=key)
            assert metadata

            # Check if local file exists
            if os.path.exists(local_file_path):
                # Get local file stats
                local_file_stat = os.stat(local_file_path)
                local_file_size = local_file_stat.st_size
                local_file_mtime = local_file_stat.st_mtime

                # Convert S3 object last modified to timestamp
                s3_last_modified = metadata['LastModified'].replace(tzinfo=timezone.utc).timestamp()

                # Compare file sizes and modification timestamps
                if local_file_size == metadata['ContentLength'] and local_file_mtime >= s3_last_modified:
                    debug_print("Local file is up to date. No download needed.")
                    return True
                else:
                    debug_print("Remote is newer or sizes differ")
                    pass
            else:
                debug_print("Local file does not exist.")
                pass
            # Download the file if size or timestamp differ, or if local file does not exist
            debug_print(f"Attemting to download {s3_uri} to {local_file_path}")
            s3_client.download_file(bucket_name, key, local_file_path)
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == "404":
                debug_print(f"{s3_uri} does not exist.", file=sys.stderr)
                return False
            else:
                raise

    @staticmethod
    def _remote_dir_exists(s3_uri: str) -> bool:
        bucket_name, prefix = S3RemoteCheckpointSynchronizer.parse_s3_uri(s3_uri)
        s3_client = boto3.client('s3')
        # exactly one trailing slash,no more, no less
        prefix_with_slash = prefix.rstrip('/') + '/'
        # retrieve some keys from the bucket that begin with the prefix
        response = s3_client.list_objects_v2(
            Bucket=bucket_name,
            Prefix=prefix_with_slash,
        )
        debug_print(f'response from list_objects_v2 is {response.__repr__()}')
        return bool('KeyCount' in response and response['KeyCount'] > 0)

S3RemoteCheckpointSynchronizer.register_thyself()


class CheckpointRegistry():
    output_dir: str
    remote_synchronizer: Union[RemoteCheckpointSynchronizer,None]
    _registry: Dict[int,CheckpointInfo]  # it is necessary to have string keys because of json serialization
    upload_in_progress: Union[RemoteCheckpointSynchronizer.Task,None]

    def __init__(self, output_dir: str, remote_synchronizer: Union[RemoteCheckpointSynchronizer,None] = None):
        self.output_dir = output_dir
        self.remote_synchronizer = remote_synchronizer
        if self.remote_synchronizer:
            assert self.output_dir == self.remote_synchronizer.local_output_dir
        self.upload_in_progress = None
        # ensure output_dir exists
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        self._load()

    def registry_path(self):
        return f"{self.output_dir}/{REGISTRY_FILENAME}"

    def save(self):
        # save _registry safely as json
        realfilename = self.registry_path()
        tempfilename = f"{self.registry_path()}.tmp"
        with open(tempfilename, 'w') as f:
            dumpable = {}
            for steps,checkpoint_info in self._registry.items():
                dumpable[str(steps)] = checkpoint_info.mini_dict()
            json.dump(dumpable, f)
        os.rename(tempfilename, realfilename)
        if self.remote_synchronizer:
            if self.upload_in_progress:
                self.upload_in_progress.wait_for_it()
                self.upload_in_progress = None
            self.upload_in_progress = self.remote_synchronizer.upload_all_async()

    def _load(self) -> None:
        registry_path = self.registry_path()

        if self.remote_synchronizer and not os.path.exists(registry_path):
            self.remote_synchronizer.download_one_sync(REGISTRY_FILENAME)

        loaded_registry:Dict[int,CheckpointInfo] = {}
        if os.path.exists(registry_path):
            with open(registry_path, 'r') as f:
                loadable = json.load(f)
                for steps_str,mini_dict in loadable.items():
                    loaded_registry[int(steps_str)] = CheckpointInfo.from_mini_dict(output_dir=self.output_dir, mini_dict = mini_dict)
        self._registry = loaded_registry

    def add_checkpoint(self, global_step: int, checkpoint: str, segment_number: Union[int,None] = None):
        checkpoint_info= CheckpointInfo(output_dir=self.output_dir,
                                        checkpoint_name= checkpoint,
                                        global_step=global_step ,
                                        segment_number= segment_number)
        self._registry[global_step] = checkpoint_info
        self.save()

    def get_checkpoint_for_step(self, global_step: int) -> Union[CheckpointInfo,None]:
        supposed_checkpoint_info: Union[CheckpointInfo,None] = self._registry.get(global_step,None)
        if self.remote_synchronizer and supposed_checkpoint_info and not supposed_checkpoint_info.exists():
            checkpoint_name = supposed_checkpoint_info.checkpoint_name
            debug_print(f'about to download_one_sync({checkpoint_name})')

            self.remote_synchronizer.download_one_sync(checkpoint_name)
        if supposed_checkpoint_info and supposed_checkpoint_info.exists():
            return supposed_checkpoint_info
        else:
            return None

    def latest_step(self) -> Union[int,None]:
        if self._registry.keys():
            return max(k for k in self._registry.keys())
        else:
            return None

    def finish_up(self):
        if self.remote_synchronizer:
            self.remote_synchronizer.finish_up()
            self.upload_in_progress = None




from setuptools import setup, find_packages

setup(
    name='s3datasets',
    python_requires='>=3.6',
    description = "System to ease incremental training of a Huggingface transformer model from a large S3-based dataset",
    long_description="This Python module provides tools to enhance Huggingface Trainer to make it feasible to train transformers on very large remote datasets that are impractical to just download en-masse for every training session or to complete all of the training on a single server instance.  It makes it practical to use cloud providers like vast.ai or salad.ai as a fine-tuning platform for large language models.",
    version='0.0.3a1',
    package_dir={'': 'src'},  # This tells setuptools where to find packages
    packages=find_packages(where='src'),
    install_requires=[
        's3datasets',
        'dataclasses',
        'typing; python_version<"3.5"',
        'peft'
    ],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ]

)





#!/usr/bin/env python

import os
import warnings
from transformers import Trainer, TrainerCallback, TrainingArguments, TrainerState, TrainerControl
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR
from typing import Union
from tokengenerators import TextDS2TokensGenerator, DSGeneratorCursor
from .checkpoint_registry import CheckpointRegistry, CheckpointInfo
import logging

default_logger = logging.getLogger(__name__)

DATASET_CURSOR_FILENAME='dataset_cursor.json'


class CheckpointManager(TrainerCallback):
      """
      Adds a callback to a huggingface transformers.Trainer to watch for checkpoints being
      made and record the cursor of the generator behind the training dataset so that
      training resumption is feasible on huge iterable datasets.

      Also ensures that the generator cursor is properly initialized when training is resumed.

      usage:
          checkpoint_callback = CheckpointManager(trainer,checkpoint_registry, dataset_generator, starting_checkpoint_info, stop_after_steps = 10000, save_on_stop = True, my_logger)
          trainer.train()
      """
      trainer: Trainer
      dataset_generator: Union[TextDS2TokensGenerator,None]
      stop_after_steps: Union[int, None]
      save_on_stop: bool
      steps_seen: int
      checkpoint_registry: Union[CheckpointRegistry,None]
      starting_checkpoint_info: Union[CheckpointInfo,None]
      current_dataset_segment_number: int
      logger: logging.Logger
      trainer_state_at_save: Union[dict,None]
      max_steps_passed: bool

      def __init__(self,
                   trainer: Trainer,
                   checkpoint_registry : Union[CheckpointRegistry,None] = None,
                   dataset_generator: Union[TextDS2TokensGenerator,None] = None,
                   starting_checkpoint_info: Union[CheckpointInfo,None] = None,
                   stop_after_steps: Union[int,None] = None,
                   save_on_stop: bool = True,
                   logger: logging.Logger = default_logger
                   ):
          super()
          self.trainer = trainer
          self.dataset_generator = dataset_generator
          self.checkpoint_registry = checkpoint_registry
          self.starting_checkpoint_info = starting_checkpoint_info
          self.stop_after_steps = stop_after_steps
          self.save_on_stop = save_on_stop
          self.logger = logger
          self.steps_seen = 0
          self._last_save_was_at_step = None
          self.trainer_state_at_save = None
          self.current_dataset_segment_number = self.starting_checkpoint_info.segment_number if (self.starting_checkpoint_info and self.starting_checkpoint_info.segment_number) else 1
          self.max_steps_passed = False
          self.trainer.add_callback(self)

      def on_train_begin(self,
                         args: TrainingArguments,
                         state: TrainerState, control: TrainerControl, **kwargs ):
          if args or state or control or kwargs: # silence unused var warnings
              pass

          checkpoint = self.starting_checkpoint_info.path() if self.starting_checkpoint_info else None
          self.logger.debug(f"CheckpointManager on_train_begin with checkpoint {checkpoint}")
          if checkpoint and self.dataset_generator:
              checkpoint_cursor  = CheckpointManager.get_checkpoint_dataset_cursor(checkpoint)
              if checkpoint_cursor:
                  self.logger.debug(f"restoring cursor for checkpoint {checkpoint}")
                  self.dataset_generator.set_cursor(checkpoint_cursor)
              else:
                  self.logger.warn(f"checkpoint {checkpoint} has no dataset cursor")
          if self.dataset_generator:
              self.logger.info(f"initial cursor is segment {self.current_dataset_segment_number}, {self.dataset_generator.get_cursor().to_dict()}")
          else:
              self.logger.debug(f"CheckpointManager is unaware of a dataset_generator")

          if state.global_step >= args.max_steps:
              self.max_steps_passed = True

      def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs ):
          if args or state or kwargs: # suppress unused param warnings
              pass
          self.logger.debug(f"about to do local step {self.steps_seen+1} and dataset cursor is segment {self.current_dataset_segment_number}, {self.dataset_generator.get_cursor().to_dict().__repr__() if self.dataset_generator else 'not defined'}")
          # Why the heck do I have to do this?  It's insane.  This should be the default behavior of Trainer anyway.
          # Instead, if it runs out of training data, it raises an error
          if self.dataset_generator and self.dataset_generator.exhausted:
              self.logger.debug("dataset_generator is exhausted. Stopping training.")
              control.should_training_stop = True

          if control.should_training_stop and self.save_on_stop:
              self.logger.debug("Trying to trigger a save_on_stop")
              control.should_save = True

      def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs ):
          if args or state or kwargs: # suppress unused param warnings
              pass
          self.steps_seen += 1
          self.logger.debug(f"just did local step {self.steps_seen} and dataset cursor is segment {self.current_dataset_segment_number}, {self.dataset_generator.get_cursor().to_dict().__repr__() if self.dataset_generator else 'not defined'}")
          if self.stop_after_steps and self.steps_seen >= self.stop_after_steps:
              self.logger.debug(f"Terminating training loop after {self.stop_after_steps} steps")
              control.should_training_stop = True

          if state.global_step >= args.max_steps:
              self.max_steps_passed = True

          # Why the heck do I have to do this?  It's insane.  This should be the default behavior of Trainer anyway.
          # Instead, if it runs out of training data, it raises an error
          if self.dataset_generator and self.dataset_generator.exhausted:
              control.should_training_stop = True

          if control.should_training_stop and self.save_on_stop:
              self.logger.debug("Trying to trigger a save_on_stop")
              control.should_save = True

      def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs ):
          if args and state and control and kwargs:
              pass
          self.logger.debug(f"detected a checkpoint save at step {state.global_step}")
          checkpoint_dir = f"{PREFIX_CHECKPOINT_DIR}-{state.global_step}"
          checkpoint_path = os.path.join(args.output_dir,checkpoint_dir)
          self._last_save_was_at_step = state.global_step
          # if something exists at checkpoint_path...
          if os.path.exists(checkpoint_path):
              global_step = state.global_step
              # make a deep copy of state for debugging purposes
              self.trainer_state_at_save = state.__dict__.copy()
              if self.dataset_generator and self.dataset_generator.exhausted:
                  self.current_dataset_segment_number += 1
                  self.logger.debug(f"Current dataset exhausted, advancing segment number to {self.current_dataset_segment_number}")
                  if self.checkpoint_registry:
                      self.dataset_generator.set_cursor(DSGeneratorCursor(0,0))
              self._save_dataset_cursor(checkpoint_path)
              if self.checkpoint_registry:
                  self.checkpoint_registry.add_checkpoint(global_step=global_step,
                                                          segment_number = self.current_dataset_segment_number,
                                                          checkpoint = checkpoint_dir)
          else:
              raise RuntimeError(f"expected checkpoint_path {checkpoint_path} does not exist")

      def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs ):
          if self.save_on_stop and ((not self._last_save_was_at_step) or self._last_save_was_at_step < state.global_step):
              self.trainer._save_checkpoint(self.trainer.model,trial=None,metrics=None)
              self.on_save(args,state,control,kwargs=kwargs)

      @staticmethod
      def checkpoint_dataset_cursor_filepath(checkpoint_path: str):
          return os.path.join(checkpoint_path,DATASET_CURSOR_FILENAME)

      @staticmethod
      def get_checkpoint_dataset_cursor(checkpoint_path: str) -> Union[DSGeneratorCursor,None]:
          cursor_file_path = CheckpointManager.checkpoint_dataset_cursor_filepath(checkpoint_path)
          if os.path.exists(cursor_file_path):
              return DSGeneratorCursor.from_file_path(cursor_file_path)
          else:
              return None
 
      def _save_dataset_cursor(self,checkpoint_path: str) -> bool:
          if self.dataset_generator:
              try:
                  cursor = self.dataset_generator.get_cursor()
              except AttributeError:
                  warnings.warn("tried to save dataset cursor in checkpoint but generator does not support cursors")
                  return False
              cursor_file_path = self.__class__.checkpoint_dataset_cursor_filepath(checkpoint_path)
              cursor.save_to_file_path(cursor_file_path)
              return True
          else:
              return False


